from magnet.utils.globals import _f
from magnet.utils.huggingface import InferenceAPI
from magnet.utils.local import LocalInference
from magnet.utils.prompts import *
from magnet.utils.data_classes import *
import requests, json

class Generate:
    def __init__(self, server: str = None, field = None, hf_token: str = None):
        """
        Initializes the Generate class.

        Args:
            server (str): The URL of the server to be used for generating the response. Default is None.
            field: Placeholder field that can be used for future implementation.
            hf_token (str): The Hugging Face token to be used for authentication when using the local inference API. Default is None.
        """
        self.server = server if not hf_token else None
        self.field = field
        self.token = hf_token

    async def on(self):
        """
        Placeholder method that can be used for future implementation.
        """
        if self.field:
            pass # todo

    async def ask(self
                  , m: str = "mistralai/Mistral-7B-Instruct-v0.1"
                  , q: str = "What is your itinerary?"
                  , t: float = 1.0
                  , n: int = 8096
                  , p: str = "qa_ref"
                  , cb: object = None
                  , docs: list = []
                  , v: bool = False
                ):
        """
        Generates a response based on a given prompt using a language model.

        Args:
            m (str): The model name or identifier to be used for generating the response. Default is "mistralai/Mistral-7B-Instruct-v0.1".
            q (str): The question or prompt for which a response is to be generated. Default is "What is your itinerary?".
            t (float): The temperature parameter controlling the randomness of the generated response. Default is 1.0.
            n (int): The maximum number of new tokens to be generated in the response. Default is 8096.
            p (str): The type of prompt to be used for generating the response. Default is "qa_ref".
            cb (object): An optional callback function to be executed with the generated response. Default is None.
            docs (list): A list of additional context or documents to be used for generating the response. Default is an empty list.
            v (bool): A flag indicating whether to use the server for generating the response. Default is False.

        Returns:
            str: The generated response.

        Raises:
            Exception: If an error occurs during the execution of the method.
        """
        prompt = getattr(globals()['Prompts'](), p)(docs,q)
        _f('warn', '(p + q + d) > n') if len(prompt) > n else None
        payload = json.dumps({
            "model": m,
            "prompt": prompt,
            "inputs": prompt,
            "parameters": {
                "max_new_tokens": n
                , "temperature": t,
            }
        })
        headers = {
            'Content-Type': 'application/json'
        }
        if self.token:
            llm = InferenceAPI(self.token)
            response = llm.invoke(payload)
            if not isinstance(response, list) and 'error' in response.keys():
                return _f('fatal', response['error'])
            else:
                response = response[0]['generated_text'].split(json.loads(payload)['prompt'])[1]
        elif v:
            response = requests.request(
                "POST", f"{self.server}/v1/completions", headers=headers, data=payload).text
        else:
            llm = LocalInference(model=m)
            response = llm.invoke(payload)
        if self.field:
            payload = GeneratedPayload(
                        query=q
                        , prompt=prompt
                        , context=docs
                        , result=response
                        , model=m
                    )
            try:
                await self.field.pulse(payload)
            except Exception as e:
                _f('fatal',e)
        if cb:
            cb(response)
        else:
            return response



import nats, json
from magnet.utils.globals import _f
from dataclasses import asdict
from nats.errors import TimeoutError
from magnet.utils.data_classes import *
from nats.js.api import StreamConfig
import xxhash

x = xxhash

class Charge:
    """
    The `Charge` class is responsible for connecting to a NATS server, managing streams and categories, and publishing data to the server.

    Args:
        server (str): The NATS server URL.

    Attributes:
        server (str): The NATS server URL.
        category (str): The current category.
        stream (str): The current stream.
        nc: The NATS connection object.
        js: The JetStream API object.
    """

    def __init__(self, server):
        self.server = server

    async def on(self, category: str = 'no_category', stream: str = 'documents', create: bool = False):
        """
        Connects to the NATS server, creates a stream and category if they don't exist, and prints a success message.

        Args:
            category (str, optional): The category to connect to. Defaults to 'no_category'.
            stream (str, optional): The stream to connect to. Defaults to 'documents'.
            create (bool, optional): Whether to create the stream and category if they don't exist. Defaults to False.
        """
        self.category = category
        self.stream = stream
        try:
            nc = await nats.connect(f'nats://{self.server}:4222')
            self.nc = nc
            self.js = self.nc.jetstream()
            self.js.purge_stream
            streams = await self.js.streams_info()
            if self.stream not in [x.config.name for x in streams] or self.category not in sum([x.config.subjects for x in streams], []):
                try:
                    if self.stream not in [x.config.name for x in streams]:
                        _f("wait", f'creating {self.stream}') \
                        , await self.js.add_stream(name=self.stream, subjects=[self.category]) \
                            if create else _f("warn", f"couldn't create {stream} on {self.server}")
                        streams = await self.js.streams_info()
                    if self.category not in sum([x.config.subjects for x in streams if x.config.name == self.stream], []):
                        subjects = sum([x.config.subjects for x in streams if x.config.name == self.stream], [])
                        subjects.append(self.category)
                        await self.js.update_stream(StreamConfig(
                            name = self.stream
                            , subjects = subjects
                        ))
                        _f("success", f'created [{self.category}] on\n🛰️ stream: {self.stream}')
                except Exception as e:
                    _f('fatal', f"couldn't create {stream} on {self.server}\n{e}")
        except TimeoutError:
            _f('fatal', f'could not connect to {self.server}')
        _f("success", f'connected to [{self.category}] on\n🛰️ stream: {self.stream}')

    async def off(self):
        """
        Disconnects from the NATS server and prints a warning message.
        """
        await self.nc.drain()
        _f('warn', f'disconnected from {self.server}')

    async def pulse(self, payload):
        """
        Publishes data to the NATS server using the specified category and payload.

        Args:
            payload (dict): The data to be published.
        """
        try:
            bytes_ = json.dumps(asdict(payload), separators=(', ', ':')).encode('utf-8')
        except Exception as e:
            _f('fatal', f'invalid JSON\n{e}')
        try:
            _hash = x.xxh64(bytes_).hexdigest()
            await self.js.publish(
                self.category
                , bytes_
                , headers={
                    "Nats-Msg-Id": _hash
                }
            )
            await self.js.publish(self.category, bytes_, headers={"Nats-Msg-Id":_hash})
        except Exception as e:
            _f('fatal', f'could not send data to {self.server}\n{e}')

    async def emp(self, name=None):
        """
        Deletes the specified stream if the name matches the current stream, or prints an error message if the name doesn't match or the stream doesn't exist.

        Args:
            name (str, optional): The name of the stream to delete. Defaults to None.
        """
        if name and name==self.stream:
            await self.js.delete_stream(name=self.stream)
            _f('warn', f'{self.stream} stream deleted')
        else:
            _f('fatal', "name doesn't match the stream or stream doesn't exist")

    async def reset(self, name=None):
        """
        Purges the specified category if the name matches the current category, or prints an error message if the name doesn't match or the category doesn't exist.

        Args:
            name (str, optional): The name of the category to purge. Defaults to None.
        """
        if name and name==self.category:
            await self.js.purge_stream(name=self.stream, subject=self.category)
            _f('warn', f'{self.category} category deleted')
        else:
            _f('fatal', "name doesn't match the stream category or category doesn't exist")

class Resonator:
    """
    The `Resonator` class is responsible for connecting to a NATS server, subscribing to a specific category in a stream, and consuming messages from that category.

    Args:
        server (str): The address of the NATS server.

    Attributes:
        server (str): The address of the NATS server.
        category (str): The category to subscribe to.
        stream (str): The stream to subscribe to.
        session (str): The session name for durable subscriptions.
        nc (nats.aio.client.Client): The NATS client.
        js (nats.aio.client.JetStream): The JetStream client.
        sub (nats.aio.client.Subscription): The subscription.

    Methods:
        __init__(self, server: str)
        on(self, category: str = 'no_category', stream: str = 'documents', cb=print, session='magnet') -> None
        info(self, session: str = None) -> None
        off(self) -> None
    """

    def __init__(self, server: str):
        """
        Initializes the `Resonator` class with the NATS server address.

        Args:
            server (str): The address of the NATS server.
        """
        self.server = server

    async def on(self, category: str = 'no_category', stream: str = 'documents', session='magnet'):
        """
        Connects to the NATS server, subscribes to a specific category in a stream, and consumes messages from that category.

        Args:
            category (str, optional): The category to subscribe to. Defaults to 'no_category'.
            stream (str, optional): The stream to subscribe to. Defaults to 'documents'.
            session (str, optional): The session name for durable subscriptions. Defaults to 'magnet'.

        Returns:
            None

        Raises:
            TimeoutError: If there is a timeout error while connecting to the NATS server.
            Exception: If there is an error in consuming the message or processing the callback function.
        """
        self.category = category
        self.stream = stream
        self.session = session
        _f('wait',f'connecting to {self.server}')
        try:
            self.nc = await nats.connect(f'nats://{self.server}:4222')
            self.js = self.nc.jetstream()
            self.sub = await self.js.subscribe(self.category, durable=self.session)
            _f("success", f'connected to {self.server}')
        except TimeoutError:
            _f("fatal", f'could not connect to {self.server}')
    async def listen(self, cb=print):
        """
        Consume messages from a specific category in a stream.

        Args:
            cb (function, optional): The callback function to process the received messages. Defaults to `print`.

        Returns:
            None

        Raises:
            Exception: If there is an error in consuming the message or processing the callback function.
        """
        _f("info", f'consuming delta from [{self.category}] on\n🛰️ stream: {self.stream}\n🧲 session: "{self.session}"')
        while True:
            try:
                msg = await self.sub.next_msg(timeout=60)
                payload = Payload(**json.loads(msg.data))
                try:
                    await cb(payload, msg)
                except Exception as e:
                    _f("warn", f'retrying connection to {self.server}')
            except Exception as e:
                _f('fatal','invalid JSON')
    async def info(self, session: str = None):
        """
        Retrieves information about a consumer in a JetStream stream.

        :param session: A string representing the session name of the consumer. If not provided, information about all consumers in the stream will be retrieved.
        :return: None
        """
        jsm = await self.js.consumer_info(stream=self.stream, consumer=session)
        _f('info',json.dumps(jsm.config.__dict__, indent=2))
    async def off(self):
        """
        Unsubscribes from the category and stream and disconnects from the NATS server.

        :return: None
        """
        await self.sub.unsubscribe()
        _f('warn', f'unsubscribed from {self.stream}')
        await self.nc.drain()
        _f('warn', f'disconnected from {self.server}')


import requests
import json

class InferenceAPI:
    """
    A class that provides a convenient way to make HTTP POST requests to an inference API endpoint.

    Attributes:
        token (str): A string representing the token used for authorization.
    """

    def __init__(self, token):
        """
        Initializes the InferenceAPI class with a token.

        Args:
            token (str): A string representing the token used for authorization.
        """
        self.token = token
    
    def invoke(self, payload):
        """
        Makes an HTTP POST request to an inference API endpoint and returns the response.

        Args:
            payload (str): A JSON string representing the payload to be sent to the inference API. It should contain the model name and input data.

        Returns:
            str: A JSON string representing the response from the inference API.
        """
        payload = json.loads(payload)
        headers = {"Authorization": f"Bearer {self.token}"}
        response = requests.post(f"https://api-inference.huggingface.co/models/{payload['model']}", headers=headers, json=payload)
        return response.json()


import pandas as pd
import os
from magnet.utils.globals import _f, Utils
from tqdm import tqdm
from magnet.utils.data_classes import *

class Processor:
    """
    The `Processor` class is responsible for loading, processing, and saving data. It provides methods for loading data from different file formats, splitting the data into smaller chunks, and saving the processed data to a file.

    Example Usage:
        # Initialize the Processor class object
        processor = Processor()

        # Load data into the processor
        processor.load(raw_data, text_column="clean", id_column="id")

        # Process and split the loaded data into smaller chunks
        await processor.process(path="processed_data.csv", splitter=None, nlp=True)

        # Save the processed data to a file
        processor.save(filename="processed_data.csv", raw=processor.df)

    Main functionalities:
        - Loading data from different file formats (csv, json, xlsx, parquet) or a pandas DataFrame
        - Splitting the loaded data into smaller chunks either by sentences using natural language processing (NLP) or by a fixed character length
        - Saving the processed data to a file in different formats (csv, json, xlsx, parquet)

    Methods:
        - __init__(): Initializes the Processor class object and sets the df attribute to None and creates an instance of the Utils class.
        - save(filename: str = None, raw: pd.DataFrame = None): Saves the pandas DataFrame to a file with the specified filename and file format.
        - load(raw: str | pd.DataFrame = None, text_column: str = "clean", id_column: str = 'id'): Loads data into the df attribute of the Processor class.
        - process(path: str = None, splitter: any = None, nlp=True): Processes and splits the loaded data into smaller chunks.
        - default_splitter(data, window_size=768, overlap=76, nlp=True): Splits the given input data into smaller chunks either by sentences or by a fixed character length.

    Fields:
        - df: A pandas DataFrame that stores the loaded data.
        - utils: An instance of the Utils class that provides utility functions.
    """
    def __init__(self):
        self.df = None
        self.utils = Utils()
        
    def save(self, filename: str = None, raw: pd.DataFrame = None):
        """
        Save the pandas DataFrame to a file with the specified filename and file format.

        Args:
            filename (str): The name of the file to save the data to.
            raw (pd.DataFrame): The pandas DataFrame containing the data to be saved.

        Raises:
            ValueError: If the data format is unsupported or an error occurs during the saving process.

        Returns:
            None: If an error occurs during the saving process.
            str: Success message if the data is successfully saved to the specified file.
        """
        try:
            file_extension = os.path.splitext(filename)[-1]
            file_handlers = {
                ".csv": raw.to_csv,
                ".json": raw.to_json,
                ".xlsx": raw.to_excel,
                ".parquet": raw.to_parquet,
            }
            if file_extension in file_handlers:
                file_handlers[file_extension](filename)
                _f("success", f"saved - {filename}")
            else:
                _f("fatal", "unsupported data")
        except Exception as e:
            _f("fatal", e)
    def load(self, raw: str | pd.DataFrame = None, text_column: str = "clean", id_column: str = 'id'):
        """
        Load data into the df attribute of the Processor class.

        Args:
            raw (str | pd.DataFrame): The input data to be loaded. It can be either a file path (str) or a pandas DataFrame.
            text_column (str): The name of the column in the input data that contains the text data. Default is "clean".
            id_column (str): The name of the column in the input data that contains the unique identifier for each data entry. Default is "id".

        Raises:
            ValueError: If the file extension is not supported or an exception occurs during the loading process.

        Returns:
            None: If an error occurs during the loading process.
            str: Success message if the data is successfully loaded.
        """
        self.id_column = id_column
        self.text_column = text_column
        try:
            if isinstance(raw, str):
                raw_data_dir = raw
                file_extension = os.path.splitext(raw)[-1]
                file_handlers = {
                    ".csv": pd.read_csv,
                    ".json": pd.read_json,
                    ".xlsx": pd.read_excel,
                    ".parquet": pd.read_parquet,
                }
                if file_extension in file_handlers:
                    _f("wait", f"loading - {raw_data_dir}")
                    self.df = file_handlers[file_extension](raw_data_dir)
                    _f("success", f"loaded - {raw_data_dir}")
                    return f"Data successfully loaded from {raw_data_dir}"
                else:
                    raise ValueError("Unsupported file type")
            elif isinstance(raw, pd.DataFrame):
                self.df = raw
                _f("success", f"loaded - {raw}")
                return f"Data successfully loaded from DataFrame"
            else:
                raise ValueError("Data type not in [csv, json, xlsx, parquet, pd.DataFrame]")
        except Exception as e:
            raise ValueError(str(e))
    async def process(self, path: str = None, splitter: any = None, nlp=True):
        """
        Process and split the loaded data into smaller chunks.

        Args:
            path (str): The path to save the processed data.
            splitter (function, optional): A custom function to split the text into chunks. If not provided, the default splitter function in the `Processor` class will be used.
            nlp (bool, optional): A flag indicating whether to split the data into sentences using natural language processing (NLP) or by a fixed character length.

        Returns:
            None: If there is no data loaded or if an error occurs during the processing.
            The processed data saved to the specified file path.
        """
        self.df = self.df.dropna()
        if self.df is not None:
            try:
                _f("wait", f"get coffee or tea - {len(self.df)} processing...")
                sentence_splitter = self.default_splitter if splitter is None else splitter
                chunks = []
                knowledge_base = pd.DataFrame()
                tqdm.pandas()
                self.df["chunks"] = self.df[self.text_column].progress_apply(
                    lambda x: [
                        str(s) for s in sentence_splitter(self.utils.normalize_text(x), nlp=nlp)
                    ]
                )
                for i in tqdm(range(len(self.df))):
                    for c in self.df['chunks'].iloc[i]:
                        d = self.df[self.id_column].iloc[i]
                        chunks.append((d, c))
                knowledge_base['id'] = [c[0] for c in chunks]
                knowledge_base['chunks'] = [c[1] for c in chunks]

                self.df = knowledge_base
                _f('wait', f'saving to {path}')
                self.save(path, self.df)
                return
            except Exception as e:
                _f("fatal", e)
        else:
            return _f("fatal", "no data loaded!")
        
    async def create_charge(self, field=None):
        """
        Process and send data to a field.

        Args:
            field (optional): The field object to which the data will be sent.

        Returns:
            None

        Raises:
            ValueError: If no data is loaded.
            ValueError: If no field is loaded.

        Example Usage:
            # Initialize the Processor class object
            processor = Processor()

            # Load data into the processor
            processor.load(raw_data, text_column="clean", id_column="id")

            # Initialize the field object
            field = Field()

            # Create charges and send data to the field
            await processor.create_charge(field)
        """
        if self.df is not None and field is not None:
            self.field = field
            for i in tqdm(range(len(self.df))):
                d = self.df[self.id_column].iloc[i]
                c = self.df[self.text_column].iloc[i]
                payload = Payload(
                    document=d,
                    text=c
                )
                if self.field:
                    await self.field.pulse(payload)
                else:
                    raise ValueError('No field initialized')
        else:
            if not self.df:
                raise ValueError('No data loaded!')
            else:
                raise ValueError('No field loaded!')
    
    def default_splitter(self, data, window_size=768, overlap=76, nlp=True):
        """
        Splits the given input data into smaller chunks either by sentences or by a fixed character length.

        Args:
            data (str): The input data to be split.
            window_size (int, optional): The size of each chunk when splitting by a fixed character length. Defaults to 768.
            overlap (int, optional): The number of characters to overlap between each chunk when splitting by a fixed character length. Defaults to 76.
            nlp (bool, optional): A flag indicating whether to split the data into sentences using natural language processing (NLP) or by a fixed character length. Defaults to True.

        Returns:
            list: If `nlp` is True, returns a list of sentences extracted from the input data. If `nlp` is False, returns a list of chunks obtained by splitting the input data into fixed-sized chunks.
        """
        if nlp:
            self.utils.nlp.max_length = len(data) + 100
            sentences = [str(x) for x in self.utils.nlp(data).sents]
            return sentences
        else:
            # Perform chunked splitting by a fixed character length
            chunks = []
            start_char_idx = 0
            while start_char_idx < len(data):
                end_char_idx = start_char_idx + window_size
                chunk = data[start_char_idx:end_char_idx]
                chunks.append(chunk)
                start_char_idx += (window_size - overlap)
            return chunks


class Prompts:
    """
    The `Prompts` class is used to generate a formatted prompt for generating an answer based on a given question and a list of documents.
    """

    def __init__(self):
        """
        Initializes an instance of the `Prompts` class.
        """
        return

    def qa_ref(self, docs: list = [], q: str = None) -> str:
        """
        Generates a formatted prompt for generating an answer based on the given documents and question.

        Args:
            docs (list): A list of dictionaries representing the documents. Each dictionary should have a "title" key and a "content" key.
            q (str): The question for which an answer is to be generated.

        Returns:
            str: A formatted string representing a prompt for generating an answer based on the given documents and question.
        """
        docs = '\n'.join([f"Document[name={d}]\ {t}\n\n" for (t, d, c) in [tuple(x.values()) for x in docs]])
        return """Create a concise and informative answer (no more than 200 words) for a given question based solely on the given documents. You must only use information from the given documents. Use an unbiased and journalistic tone. Do not repeat text. Cite the documents using Document[name] notation. If multiple documents contain the answer, cite those documents like ‘as stated in Document[name], Document[name], etc.’. If the documents do not contain the answer to the question, say that ‘answering is not possible given the available information.’
    [DOCUMENTS]
    Question: [QUERY]; Answer: """ \
                .replace('[DOCUMENTS]', docs) \
                .replace('[QUERY]', q)


from pymilvus import connections, utility, FieldSchema, CollectionSchema, DataType, Collection
from magnet.utils.globals import _f
import random, array

class MilvusDB:
    """
    A class that provides a high-level interface for interacting with a Milvus database.

    Args:
        config (dict): A configuration dictionary that contains the necessary parameters for connecting to the Milvus server and creating a collection.

    Attributes:
        config (dict): A configuration dictionary that contains the necessary parameters for connecting to the Milvus server and creating a collection.
        fields (list): A list of `FieldSchema` objects that define the fields of the collection.
        schema (CollectionSchema): A `CollectionSchema` object that represents the schema of the collection.
        index_params (dict): A dictionary that contains the parameters for creating an index on the collection.
        connection: The connection object to the Milvus server.
        collection: The collection object in MilvusDB.

    """

    def __init__(self, config):
        self.config = config
        self.fields = [
            FieldSchema(name='id', dtype=DataType.INT64, is_primary=True, auto_id=True),
            FieldSchema(name='document', dtype=DataType.VARCHAR, max_length=4096),
            FieldSchema(name='text', dtype=DataType.VARCHAR, max_length=65535),
            FieldSchema(name='embedding', dtype=DataType.FLOAT_VECTOR, dim=self.config['DIMENSION'])
        ]
        self.schema = CollectionSchema(fields=self.fields)
        self.index_params = self.config['INDEX_PARAMS']

    def on(self):
        """
        Establishes a connection to the Milvus server and creates a collection object.

        Returns:
            None

        Raises:
            Exception: If an error occurs during the connection to the Milvus server.

        """
        try:
            self.connection = connections.connect(
                host=self.config['MILVUS_URI']
                , port=self.config['MILVUS_PORT']
                , user=self.config['MILVUS_USER']
                , password=self.config['MILVUS_PASSWORD']
                , alias='magnet'
            ) \
                if 'MILVUS_PASSWORD' in self.config and 'MILVUS_USER' in self.config \
                else connections.connect(
                    host=self.config['MILVUS_URI']
                    , port=self.config['MILVUS_PORT']
                    , alias='magnet'
                )
            self.collection = Collection(name=self.config['INDEX'], schema=self.schema)
            self.collection.load()
            _f('success', f"connected successfully to {self.config['MILVUS_URI']}")
        except Exception as e:
            _f('fatal', e)

    def off(self):
        """
        Disconnects from the Milvus server.

        Returns:
            None

        Raises:
            Exception: If an error occurs during the disconnection from the Milvus server.

        """
        try:
            self.connection = connections.disconnect(alias="magnet")
            _f('warn', f"disconnected from {self.config['MILVUS_URI']}")
        except Exception as e:
            _f('fatal', e)

    def create(self, overwrite=False):
        """
        Create a collection in MilvusDB and create an index on a specific field of the collection.

        Args:
            overwrite (bool, optional): A boolean flag indicating whether to overwrite the existing collection with the same name. Default is False.

        Returns:
            None

        Raises:
            Exception: If an error occurs during the creation of the collection or index.

        Example Usage:
            config = {
                'MILVUS_URI': 'localhost',
                'INDEX': 'my_collection',
                'DIMENSION': 128,
                'INDEX_PARAMS': {'index_type': 'IVF_FLAT', 'nlist': 100}
            }
            milvus_db = MilvusDB(config)
            milvus_db.create(overwrite=True)

        """
        if utility.has_collection(self.config['INDEX']) and overwrite:
            utility.drop_collection(self.config['INDEX'])
        try:
            self.collection = Collection(name=self.config['INDEX'], schema=self.schema)
            self.collection.create_index(field_name="embedding", index_params=self.index_params)
            _f('success', f"{self.config['INDEX']} created")
        except Exception as e:
            _f('fatal', e)

    def initialize(self, user: str = 'magnet', password: str = '33011033'):
        try:
            if not self.connection:
                return _f('fatal','no active Milvus connection') 
            else:
                _f('warn', f"initializing {self.config['MILVUS_URI']} using `root` to create '{user}'")
            utility.reset_password('root', 'Milvus', self._pw(), using='magnet')
            _f('success', f"secured root user successfully on {self.config['MILVUS_URI']}")
            utility.create_user(user, password, using='magnet')
            _f('success', f"created requested {user} on {self.config['MILVUS_URI']}")
            try:
                self.connection = connections.connect(
                    host=self.config['MILVUS_URI']
                    , port=self.config['MILVUS_PORT']
                    , user=user
                    , password=password
                )
                _f('success', 'Milvus has been initialized with your new credentials')
            except Exception as e:
                _f('fatal', e)
        except Exception as e:
            _f('fatal', e)
    
    def delete_index(self):
        """
        Deletes the index of a collection in MilvusDB.

        Returns:
            None

        """
        if utility.has_collection(self.config['INDEX']):
            utility.drop_collection(self.config['INDEX'])
            _f('warn', f"{self.config['INDEX']} deleted")

    def _pw(self):
        MAX_LEN = 24
        DIGITS = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']   
        LOCASE_CHARACTERS = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h',  
                            'i', 'j', 'k', 'm', 'n', 'o', 'p', 'q', 
                            'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 
                            'z'] 
        UPCASE_CHARACTERS = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',  
                            'I', 'J', 'K', 'M', 'N', 'O', 'P', 'Q', 
                            'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 
                            'Z'] 
        SYMBOLS = ['@', '#', '$', '%', '=', ':', '?', '.', '/', '|', '~', '>',  
                '*', '(', ')', '<'] 
        COMBINED_LIST = DIGITS + UPCASE_CHARACTERS + LOCASE_CHARACTERS + SYMBOLS 
        rand_digit = random.choice(DIGITS) 
        rand_upper = random.choice(UPCASE_CHARACTERS) 
        rand_lower = random.choice(LOCASE_CHARACTERS) 
        rand_symbol = random.choice(SYMBOLS) 
        temp_pass = rand_digit + rand_upper + rand_lower + rand_symbol 
        for x in range(MAX_LEN - 4): 
            temp_pass = temp_pass + random.choice(COMBINED_LIST)  
            temp_pass_list = array.array('u', temp_pass) 
            random.shuffle(temp_pass_list) 
        password = "" 
        for x in temp_pass_list: 
            password = password + x 
        _f('warn', f'Your Milvus `root` user password is now {password}')
        return password


from dataclasses import dataclass

@dataclass
class Payload:
    """
    Represents a payload with two main fields: text and document.

    Args:
        text (str): The text associated with the payload.
        document (str): The document associated with the payload.
    """
    text: str
    document: str

@dataclass
class GeneratedPayload:
    """
    Represents a payload generated by a system.

    Args:
        query (str): The query associated with the payload.
        prompt (str): The prompt associated with the payload.
        context (list): The context associated with the payload.
        result (str): The result generated by the system.
        model (str): The model used to generate the payload.
    """
    query: str
    prompt: str
    context: list
    result: str
    model: str

@dataclass
class EmbeddingPayload:
    """
    Represents a payload for embedding text data.

    Attributes:
        document (str): The document associated with the text data.
        embedding (list): The embedding of the text data.
        text (list): The text of the data.
        model (str): The model used for embedding the text data.
    """
    document: str
    embedding: list
    text: list
    model: str

@dataclass
class MistralArgs:
    """
    Represents a set of arguments for the Mistral model.

    Args:
        dim (int): The dimensionality of the model.
        n_layers (int): The number of layers in the model.
        head_dim (int): The dimensionality of each attention head.
        hidden_dim (int): The dimensionality of the hidden layer in the feed-forward network.
        n_heads (int): The number of attention heads.
        n_kv_heads (int): The number of attention heads used for key-value attention.
        norm_eps (float): The epsilon value used for numerical stability in layer normalization.
        vocab_size (int): The size of the vocabulary used in the model.
    """

    dim: int
    n_layers: int
    head_dim: int
    hidden_dim: int
    n_heads: int
    n_kv_heads: int
    norm_eps: float
    vocab_size: int


from sentence_transformers import SentenceTransformer
from magnet.utils.globals import _f
from magnet.utils.milvus import *
from magnet.utils.data_classes import EmbeddingPayload

class Embedder:
    """
    The Embedder class is responsible for embedding text using a pre-trained sentence transformer model and storing or sending the embeddings for further processing. It utilizes the Milvus database for storing and searching the embeddings.

    Args:
        config (dict): A dictionary containing the configuration parameters for the Embedder class, including the model name, index name, and index parameters.
        create (bool, optional): If set to True, a connection to the Milvus database will be created. Defaults to False.

    Attributes:
        config (dict): A dictionary containing the configuration parameters for the Embedder class, including the model name, index name, and index parameters.
        model (SentenceTransformer): An instance of the SentenceTransformer class from the sentence_transformers library, used for text embedding.
        db (MilvusDB): An instance of the MilvusDB class from the magnet.utils.milvus module, used for connecting to the Milvus database.

    Methods:
        embed_and_store(self, payload, verbose=False, field=None): Embeds the given payload using a pre-trained sentence transformer model and stores it in the Milvus database.
        embed_and_charge(self, payload, verbose=False, field=None): Embeds the given payload using a pre-trained sentence transformer model and sends it to a specified field for further processing.

    """

    def __init__(self, config, create=False, initialize=False):
        self.config = config
        self.model = SentenceTransformer(self.config['MODEL'])
        self.db = MilvusDB(self.config)
        self.db.on()
        if create:
            self.db.create(overwrite=True)
        if initialize:
            self.db.initialize()

    async def index(self, payload, msg, verbose=False, field=None, charge=False, instruction: str = "Represent this sentence for searching relevant passages: "):
        """
        Embeds the given payload using a pre-trained sentence transformer model and stores it in a Milvus database.

        Args:
            payload (object): An object containing the text and document attributes to be embedded and stored.
            verbose (bool, optional): A boolean indicating whether additional information should be logged during the embedding process. Defaults to False.
            field (object, optional): An object representing the field to which the encoded payload will be sent for further processing.

        Returns:
            None

        Raises:
            Exception: If an error occurs during the embedding or storing process.

        """
        if not msg:
            return _f('fatal', 'no field message to ack!')
        if field:
            self.field = field
        else:
            raise ValueError('field is required')
        try:
            _f('info', 'embedding payload') if verbose else None
            payload.embedding = self.model.encode(
                f"{instruction} {payload.text}", normalize_embeddings=True)
        except Exception as e:
            _f('fatal', e)
        else:
            await msg.in_progress()
            try:
                _f('info', 'storing payload') if verbose else None
                if not self.is_dupe(q=payload.embedding):
                    self.db.collection.insert([
                        [payload.document], [payload.text], [payload.embedding]
                    ])
                    self.db.collection.flush(collection_name_array=[
                                            self.config['INDEX']])
                    if charge:
                        payload = EmbeddingPayload(
                            model=self.config['MODEL'],
                            embedding=self.model.encode(
                                f"{instruction} {payload.text}", normalize_embeddings=True).tolist(),
                            text=payload.text,
                            document=payload.document
                        )
                        await msg.ack()
                        _f('info', f'sending payload\n{payload}') if verbose else None
                        await self.field.pulse(payload)
                    msg.ack()
            except Exception as e:
                _f('fatal', e)

    def search(self, payload, limit=100, cb=None, instruction="Represent this sentence for searching relevant passages: "):
        """
        Search for relevant passages based on a given input payload.

        Args:
            payload (str): The search query or input payload.
            limit (int, optional): The maximum number of results to return. Defaults to 100.
            cb (function, optional): A callback function to process the results. Defaults to None.

        Returns:
            list: A list of dictionaries, each containing the text, document, and distance of a relevant passage.
        """
        payload = EmbeddingPayload(
            text=payload,
            embedding=self.model.encode(
                f"{instruction} {payload}", normalize_embeddings=True),
            model=self.config['MODEL']
        )
        
        _results = self.db.collection.search(
            data=[payload.embedding],  # Embeded search value
            anns_field="embedding",  # Search across embeddings
            param=self.config['INDEX_PARAMS'],
            limit=limit,  # Limit to top_k results per search
            output_fields=['text', 'document']
        )
        results = []
        for hits_i, hits in enumerate(_results):
            for hit in hits:
                results.append({
                    'text': hit.entity.get('text'),
                    'document': hit.entity.get('document'),
                    'distance': hit.distance
                })
        if cb:
            return cb(payload.text, results)
        else:
            return results

    def info(self):
        return self.db.collection

    def disconnect(self):
        return self.db.off()

    def delete(self, name=None):
        """
        Delete an index from the Milvus database.

        Args:
            name (str, optional): The name of the index to be deleted. If not provided, no index will be deleted.

        Returns:
            None

        Raises:
            Exception: If an error occurs during the deletion process.

        Example Usage:
            config = {
                'MODEL': 'bert-base-nli-mean-tokens',
                'INDEX': 'my_index',
                'INDEX_PARAMS': {'nprobe': 16}
            }
            embedder = Embedder(config)
            embedder.delete('my_index')
        """
        if name and name == self.config['INDEX']:
            try:
                self.db.delete_index()
            except Exception as e:
                _f('fatal', e)
        else:
            _f('fatal', "name doesn't match the connection or the connection doesn't exist")
    
    def is_dupe(self, q):
        """
        Check if a given query is a duplicate in the Milvus database.

        Args:
            q (object): The query embedding to check for duplicates.

        Returns:
            bool: True if the query embedding is a duplicate in the Milvus database, False otherwise.
        """
        match = self.db.collection.search(
            data=[q]
            , anns_field = "embedding"
            , param=self.config['INDEX_PARAMS']
            , output_fields=['text', 'document']
            , limit=1
        )
        return True if sum(match[0].distances) == 0.0 else False


from setuptools import setup

setup(
    name='llm_magnet',
    version='0.1.6',
    description="the small distributed language model toolkit. fine-tune state-of-the-art LLMs anywhere, rapidly.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "torch"
        , "spacy"
        , "numpy"
        , "nats-py"
        , "sentence_transformers"
        , "pymilvus"
        , "rich"
        , "xxhash"
    ],
    url = 'https://github.com/Prismadic/magnet'
)